/* -----------------------------------------------
 * Rolladensteuerung V 1.0 von Andreas Stockburger
 * -----------------------------------------------
 *
 * Diese Rolladensteuerung steuert 6 Gruppen von Rolladen mit 6 Rolladentastern (12 Taster - je 2 fr einen Rolladen). 
 *
 * Ausgehend von den jeweils gedrckten Tasten werden die Relais bedient. Es wird pro Rolladen ein Richtungsrelais und 
 * ein Schaltrelais bedient. Die Relais werden bestmglich "geschont" es wird immer zuerst ds Richtungsrelais bedient.
 * Mit einer Zeitverzgerung wird das Schaltrelais gesschalten.
 *
 * Tastendrcke knnen durch reale Tasten (Rolladentaster) aber auch virtuellen Tasten erzeugt werden. Virtuelle Tasten
 * werden durch Doppelklick oder zeitgesteuert ausgelst. Diese  bestimmen dann die Position der jeweiligen Relais.
 *
 * Hardware:
 * ---------
 * 12 Taster sind mit PB0-PB4 und PA0-PA7 verbunden. Die Taster schalten +5V gegen Masse. D.h. Sie sind low-aktiv.
 * Ein nicht gedrckter Taster liefert "1", ein gedrckter liefert "0"
 *
 * Die Relais hngen an PD3-PD4 und PC0-PC7 und werden gegen Masse geschalten. "1" am Port fhrt zu angezogen Relais.
 *
 * An PB4 hngt eine Status-LED
 *
 * An XTAL1 ist ein 10Mhz Quarz-Oszillator angeschlossen.
 *
 * An PD0+PD1 ist die Serielle Schnittstelle mittels MAX232 beschaltet
 
 * Konfig des Controllers:
 *----------------------
 *
 * EESAVE-Fuse setzen !!!!
 *
 */

//Defines zur Anpassung an Quarze (Produktionsumgebung 10Mhz)
#define F_CPU 10000000UL // 7372800UL // Takt in Herz
#define T_VAL 50000 // 36864		// Timer1-Wert 10MHz=50000; 7,xMHz=36864 (F_CPU/T_VAL muss ganzzahlig sein)
//#define F_CPU 14745600UL // 7372800UL // Takt in Herz
//#define T_VAL 73728 // 36864		// Timer1-Wert 10MHz=50000; 7,xMHz=36864 14,x MHz=73728 (F_CPU/T_VAL muss ganzzahlig sein)
#define ISR_VAL1 F_CPU/T_VAL-1 // 10000000/50000=200 -> 199*50000+(50000-x) x=quarz_adjustment
#define BAUD 64 // 47 // 64 bei 10MHz; 47 bei 7,x MHz; 94 bei 14,x MHz

//Includes
#include <string.h>
#include <stdlib.h>
#include <util/delay.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>

//Variable fr Rolladenlaufzeit
uint8_t time_tab_total[12] = {18,18,18,18,18,18,18,18,18,18,18,18}; //Default-Werte fr Rolladenlaufzeit
uint8_t time_tab[12]     = {0,0,0,0,0,0,0,0,0,0,0,0};

//Variablen fr Eingabe
volatile uint16_t real_keys,real_keys_old,doppel_klick,real_keys_pressed;
volatile uint8_t klick_timer_tab[12] = {0,0,0,0,0,0,0,0,0,0,0,0}; // Timer-Speicher fr Doppelklick-Zeiten
volatile uint8_t doppel_klick_val=ISR_VAL1; // Default-Wert fr doppellick ~1s

//Variablen wg. Zugriff aus untersch. Funktionen
uint16_t output_ist, output_soll; // Zugriff aus main und aus der Kofig heraus
uint8_t template_nr=3; // Rolle der Steuerung: 0=UG, 1=EG, 2=OG 3=Frei konfigurierbar
uint16_t template_table[4][12] ={	{2730,1365,2730,1365,2730,1365,2730,1365,2730,1365,2730,1365}, // UG
									{2730,1365,2730,1365,2730,1365,2730,1365,2730,1365,2730,1365}, // EG
									{2730,1365,4075,4055,4075,4055,3775,3455,3775,3455,2730,1365}, // OG
									{2688,1344,2688,1344,2688,1344,0,0,10,5,10,5}	};// User-Tab

// Strings fr generelle Verwendung
char string60[]="1234567890123456789012345678901234567890123456789012345678901";
char string15[]="1234567890123456";

//Rollanmapping fr uhrzeitgesteuerten Betrieb
uint16_t zeitgest_hoch_val=2698;  // 101010101010
uint16_t zeitgest_runter_val=1349;	// 010101010101

//Variablen fr die Uhr
typedef struct
{
	uint8_t sec;
	uint8_t min;
	uint8_t hour;
	uint8_t day;
	uint8_t wday;
	uint8_t month;
	uint16_t year;
	uint8_t dls_flag; // 1=Sommerzeit im Okt. bereits zurckgestellt
	uint8_t leapyear;
} datestruct;
volatile datestruct zeit; // enthlt die aktuelle Zeit
volatile datestruct hoch_mofr, hoch_saso, runter_mofr, runter_saso; // autom. Schaltzeiten
uint8_t zeitsteuer_enable=0;

volatile uint8_t int_cnt=0; //Zhler Anzahl der fr Interrupts in einer Sekunde
volatile uint16_t quarz_adjustment=200; // Default-Wert fr Quarz-Adjustment (fr exakte Sekunde)

//Uptime-Varilen
typedef struct
{
	uint8_t sec;
	uint8_t min;
	uint8_t hour;
	uint16_t day;
} uptimestruct;
volatile uptimestruct uptime;

//Variablen fr EEPROM-Usage
uint16_t eeprom_quarz_adjustment EEMEM;
uint8_t eeprom_template_nr EEMEM;
uint8_t eeprom_doppel_klick_val EEMEM;
uint8_t eeprom_time_tab_total[12] EEMEM;

//Zeitsteuerung
datestruct eeprom_hoch_mofr EEMEM;
datestruct eeprom_runter_mofr EEMEM;
datestruct eeprom_hoch_saso EEMEM;
datestruct eeprom_runter_saso EEMEM;
uint8_t eeprom_zeitsteuer_enable EEMEM;
uint16_t eeprom_zeitgest_hoch_val EEMEM;
uint16_t eeprom_zeitgest_runter_val EEMEM;

//Mapping-Tab
uint16_t eeprom_template_table_3[12] EEMEM;

void toggle_LED(void) // Toggled die LED an PB4
{
	if (PINB & (1<<PINB4)) {
		PORTB &= ~(1<<PB4);
	}
	else {
		PORTB |=  (1<<PB4);
	}
}

void calc_zeit(void) // Berechnet Uhrzeit, Datum, Sommerzeit, Schaltjahr etc.
{
	uint8_t m_end[2][12] = {	{31,28,31,30,31,30,31,31,30,31,30,31}, // #Tage im Monat
								{31,29,31,30,31,30,31,31,30,31,30,31}	}; // s.o. mit Sommerzeit
	zeit.sec++;
	if (zeit.sec>59) { // Sekunde
		zeit.sec=0;
		zeit.min++;
		if (zeit.min>59) { // Minute
			zeit.min=0;
			zeit.hour++;
			//Sommerzeit im Mrz
			if (zeit.month==3 && zeit.wday==0 && zeit.day >= 25  &&zeit.hour==2) {
				zeit.hour++;
				zeit.dls_flag=0;
			}
			//Sommerzeit im Oktober Achtung: Durch Rckstellen der Uhr mgliche Endlosschleife daher Flag
			if (zeit.month==10 && zeit.wday==0 && zeit.day >= 25 && zeit.hour==3 && zeit.dls_flag==0) {
				zeit.hour--;
				zeit.dls_flag=1;
			}
			if (zeit.hour>23) { // Stunde
				zeit.hour=0;
				zeit.day++;
				zeit.wday++;
				if (zeit.wday>6) { // Wochentag
					zeit.wday=0;
				}
				//Schaltjahr
				if ((zeit.year%4==0)&&((zeit.year%100!=0)||(zeit.year%400==0)))
					zeit.leapyear = 1;
				else
					zeit.leapyear = 0;
				//Tag
				if (zeit.day>m_end[zeit.leapyear][zeit.month-1]) {
					zeit.day=1;
					zeit.month++;
				}
				//Jahr
				if (zeit.month>12) {
					zeit.month=1;
					zeit.year++;
				}
			}
		}
	}
}

void calc_uptime(void) // Berechnet Uptime
{
	uptime.sec++;
	if (uptime.sec>59) { // Sekunde
		uptime.sec=0;
		uptime.min++;
		if (uptime.min>59) { // Minute
			uptime.min=0;
			uptime.hour++;
			if (uptime.hour>23) { // Stunde
				uptime.hour=0;
				uptime.day++;
			}
		}
	}
}

ISR(TIMER1_COMPA_vect) // Interrupt-Routine
{
	uint8_t i;

	// fortschreiben uhr und kalender
	int_cnt++;
	if (int_cnt>ISR_VAL1-1) {
		OCR1A = T_VAL-quarz_adjustment; // abgleichen auf 1,0000000s
		if (int_cnt>ISR_VAL1) {
			OCR1A = T_VAL;
			int_cnt=0;
			calc_zeit();
			calc_uptime();
			toggle_LED();
		}
	}

	// Abfragen der Physikalischen Ports
	real_keys= (((uint16_t)PINA) << 4) | ((uint16_t)(PINB & 0x0f)); // Erstellung 12-BIT-Zahl abh. von Rolladenasten

	// Doppelklick-Erkennung
	// (1<<i) kann noch vereinfacht werden, wenn in jedem Schleifendurchlauf um 1 geschoben wird ...
	for (i=0;i<12;i++) {
		if (klick_timer_tab[i]>0) klick_timer_tab[i]--; // Prft auf Timer >0 und decremiert; entspricht "Continue"
		else doppel_klick&=~(1<<i); // lscht BIT i in Variable

		if (((real_keys&(1<<i)) == 0)&&((real_keys_old&(1<<i)) > 0)) { // Taste wurde gedrckt "fallende Flanke"
			real_keys_pressed=1; //Variable um im Wizzard den Tastendruck einfach abfragen zu knnen
			if (klick_timer_tab[i] > 0) { // falls zeit noch nicht abgelaufen dann doppelklick
				doppel_klick|= (1<<i); // doppelklick erkannt: Setze BIT i
			}
			else klick_timer_tab[i]=doppel_klick_val;// setze neue Endzeit fr Taster
		}
	}
	real_keys_old=real_keys;
}


int uart_putc(unsigned char c) // Sendet ein Zeichen an die ser. Schnittstelle
{
	while (!(UCSRA & (1<<UDRE))) {
		;    /* warten bis Senden moeglich */
	}
	UDR = c;                      /* sende Zeichen */
	return 0;
}

void uart_puttext(char array[]) // gib text auf uart aus bis \0
{
	uint8_t i=0;

	while (array[i]!='\0') {
		uart_putc(array[i]);
		i++;
	}
}

void uart_putcrlf(void) // gib cr lf aus
{
	uart_puttext("\n\r");
}

void itos(char string[], uint8_t zahl, uint8_t stellen) // Konvertiert int zu ASCII mit Anzahl Stellen und fhr. Null
{
	uint8_t zehner=0;
	uint8_t hunderter=0;

	while (zahl>99) {
		zahl-=100;
		hunderter++;
	}
	while (zahl>9) {
		zahl-=10;
		zehner++;
	}
	switch (stellen) {
	case 3:
		string[0]=hunderter+0x30;
		string[1]=zehner+0x30;
		string[2]=zahl+0x30;
		string[3]='\0';
		break;
	case 2:
		string[0]=zehner+0x30;
		string[1]=zahl+0x30;
		string[2]='\0';
		break;
	case 1:
		string[0]=zahl+0x30;
		string[1]='\0';
		break;
	}
}

void ltos(char string[], uint16_t zahl, uint8_t stellen) // Konvertiert long zu ASCII mit Anzahl Stellen und fhr. Null
{
	uint8_t zehner=0;
	uint8_t hunderter=0;
	uint8_t tausender=0;
	uint8_t zehntausender=0;

	while (zahl>9999) {
		zahl-=10000;
		zehntausender++;
	}
	while (zahl>999) {
		zahl-=1000;
		tausender++;
	}
	while (zahl>99) {
		zahl-=100;
		hunderter++;
	}
	while (zahl>9) {
		zahl-=10;
		zehner++;
	}
	switch (stellen) {
	case 5:
		string[0]=zehntausender+0x30;
		string[1]=tausender+0x30;
		string[2]=hunderter+0x30;
		string[3]=zehner+0x30;
		string[4]=zahl+0x30;
		string[5]='\0';
		break;
	case 4:
		string[0]=tausender+0x30;
		string[1]=hunderter+0x30;
		string[2]=zehner+0x30;
		string[3]=zahl+0x30;
		string[4]='\0';
		break;
	case 3:
		string[0]=hunderter+0x30;
		string[1]=zehner+0x30;
		string[2]=zahl+0x30;
		string[3]='\0';
		break;
	case 2:
		string[0]=zehner+0x30;
		string[1]=zahl+0x30;
		string[2]='\0';
		break;
	case 1:
		string[0]=zahl+0x30;
		string[1]='\0';
		break;
	}
}

void print_nibble(uint16_t inp,uint8_t nibble) // nibble 0,1 oder 2
{
	uint8_t i;
	uint16_t t=0;

	for (i=0;i<4;i++) {
		switch (nibble) {
		case 0:
			t=inp&0x0008;
			break;
		case 1:
			t=inp&0x0080;
			break;
		case 2:
			t=inp&0x0800;
			break;
		case 4:
			t=inp&0x8000; // wird nicht gebraucht
			break;
		}
		if (t==0) uart_putc('0');
		else uart_putc('1');
		inp=inp<<1;
	}
}

uint8_t decode_sequence(uint8_t neu, uint8_t alt) //Ermittelt Schonstellungen fr Relais
{
	/*
	Abh. vom Ist-Zustand und dem zu ereichenden Ziel-Zustand sind bis zu 3 Schaltschritte notwendig
	die Wahrheitstabelle zeigt die logik. Das Richtungsrelais darf nur Stromlos geschalten werden.

	I=Ist-Zustand R=Richtungselais
	Z=Zielzustand S=Schaltrelais
	S1..3= Stati mit jeweils Zeitvz. 100ms

	I  Z     S1   S2   S3  S1-Dezimal
	RS RS    RS   RS   RS
	00 00 -> 00 - 00 - 00  0
	00 01 -> 01 - 01 - 01  1
	00 10 -> 10 - 10 - 10  2
	00 11 -> 10 - 11 - 11  3
	01 00 -> 00 - 00 - 00  0
	01 01 -> 01 - 01 - 01  1
	01 10 -> 00 - 10 - 10  0
	01 11 -> 00 - 10 - 11  0
	10 00 -> 00 - 00 - 00  0
	10 01 -> 00 - 01 - 01  0
	10 10 -> 10 - 10 - 10  2
	10 11 -> 11 - 11 - 11  3
	11 00 -> 10 - 00 - 00  2
	11 01 -> 10 - 00 - 01  2
	11 10 -> 10 - 10 - 10  2
	11 11 -> 11 - 11 - 11  3

	S2 und S3 ergeben sich iterativ aus S1. Daher nur Tab fr S1
	*/
	uint8_t table[16] = {0,1,2,2,0,1,0,0,0,0,2,3,2,2,2,3}; // Tab fr Relais-Schonstellungen
	uint8_t tn,ta,ind,erg,i,j,wert;

	erg=0;
	for (i=0;i<4;i++) {
		tn=neu & 0xc0;
		ta=alt & 0xc0;
		ind=(ta>>4)+(tn>>6);
		wert=table[ind];
		for (j=3;j>i;j--) {
			wert=wert<<2;
		}
		erg=erg|wert;
		neu<<=2;
		alt<<=2;
	}
	return (erg);
}

void switch_relais(uint16_t soll , uint16_t ist) //Schaltet Relais
{
	uint8_t tc_soll,td_soll,tc_ist,td_ist,i;

	//convert 12-BIT zu 8 und 4 BIT
	tc_soll=(uint8_t)(soll>>4);
	td_soll=(uint8_t)(soll&0x0f);
	tc_ist=(uint8_t)(ist>>4);
	td_ist=(uint8_t)(ist&0x0f);

	for (i=0;i<3;i++)
	{
		tc_ist=decode_sequence(tc_soll,tc_ist);
		PORTC=tc_ist;
		td_ist=decode_sequence(td_soll,td_ist);
		PORTD=td_ist<<4;
		if (i<2) _delay_ms(125); // Schaltverzgerung der Relais xxx
	}
}

uint16_t keys2relais(uint16_t input) // mappt 2 Rolladen (4 Taster) auf 4 Relais
{
	/*
	Tabelle gilt fr 2 Rolladen (4 Relais) und 2 Rolladenschalter (4 Taster).
	Es wird dargestellt, welche	Tastenkombinationen zu welchen Relaisstellungen fhren

	R=Taster fr Runter lowaktiv (0=gedrckt)
	H=Taster fr Hoch lowaktiv (0=gedrckt)
	       R=Richtungsrelais;
		   S=Schaltrelais
	RHRH | RSRS Dezimal
	1111 | 0000 0
	1110 | 0011 3
	1101 | 0001 1
	1100 | 0000 0
	1011 | 1100 12
	1010 | 1111 15
	1001 | 1101 13
	1000 | 1100 12
	0111 | 0100 4
	0110 | 0111 7
	0101 | 0101 5
	0100 | 0100 4
	0011 | 0000 0
	0010 | 0011 3
	0001 | 0001 1
	0000 | 0000 0
	*/

	uint8_t table[16] = {0,1,3,0,4,5,7,4,12,13,15,12,0,1,3,0}; // Tab. fr Zuordnung Relais zu Tasten
	uint16_t t,output;

	t=input&0x0F00;
	t>>=8;
	output=table[t];
	output<<=4;

	t=input&0x00F0;
	t>>=4;
	output|=table[t];
	output<<=4;

	t=input&0x000F;
	output|=table[t];

	return(output);
}

uint16_t map_keys(uint16_t input)
{
	uint16_t i,output;

	output=4095;
	for (i=0;i<12;i++) {
		if ((input&(1<<i))==0) output&=template_table[template_nr][i];
	}
	return(output);
}

uint16_t startstop_time(uint16_t keys, uint16_t old_keys, uint16_t old_value) // Setzt Endzeit pro Taster
{
	uint16_t i,new_value;

	// (1<<i) kann nch vereinfacht werden, wenn in jedem Schleifendurchlauf um 1 geschoben wird ...
	new_value=old_value;
	for (i=0;i<12;i++) {
		if (((keys&(1<<i)) == 0)&&((old_keys&(1<<i)) > 0)) { // Taste wurde gedrckt "fallende Flanke"
			time_tab[i]=zeit.sec+time_tab_total[i]; // setze neue Endzeit fr Taster
			if (time_tab[i] > 59) time_tab[i]=time_tab[i]-59;// Berechne und Korrigiere berlauf
			if ((old_value&(1<<i)) == 0) new_value|=(1<<i); // setze virtuellen Taster xxx
			if ((old_value&(1<<i)) > 0) new_value&=~(1<<i); // lsche (aktiviere) virtuellen Tater
		}
	}
	return(new_value);
}

uint16_t continue_time(uint16_t old_value) // Prft auf Abschalt(zeit)bedingung pro Taster
{
	uint16_t i,new_value;

	// (1<<i) kann nch vereinfacht werden, wenn in jedem Schleifendurchlauf um 1 geschoben wird ...
	new_value=old_value;
	for (i=0;i<12;i++) {
		if	((old_value&(1<<i)) == 0) {		// ist v-taster gedrckt ?
			if (time_tab[i] == zeit.sec) new_value|=(1<<i); // prfe ob zeit in tab erreicht
		}
	}
	return(new_value);
}

uint8_t is_printable(char in)
{
	if ((in>32) && (in<127)) return(1);
	else return (0);
}

void inp_string(char text[], int8_t maxchar  ) // Zeiger auf String und maximle Anzahl von Bytes (incl. \0)
{
	uint8_t i;
	char t;

	i=0;
	t='\0';
	maxchar--; // maxchar Anzahl inclusive \0 daher 1 abziehen u, platz fr \0 zu resevieren
	while (t!='\r') {						// cr als Eingabeende
		while (!(UCSRA & (1<<RXC))) {
			;    //wartet auf Eingabe
		}
		t=UDR;
		if (t==0x08) { // Backspace
			if (i>0) { // i=0 ist kleinster wert
				i--;
				uart_putc(t); // Echo der Eingabe
			}
		}
		else {
			if ((maxchar>i)&&is_printable(t)) { // Maximale Lnge n0ch nicht erreicht ?
				text[i]=t;
				i++;
				uart_putc(t);// Echo der Eingabe nur wenn nicht cr
				text[i]='\0'; //nchstes Zeichen wird \0
			}
		}
	}
}

void inp_bits(uint16_t *zahl) // Zeiger auf long
{
	uint8_t i,maxbits;
	char t;

	t='\0';
	maxbits=12;  // Rolladensteuerung nutzt 12 BIT
	i=maxbits; // i=1->BIT0, i=2->BIT1 ...

	while (t!='\r') {						// cr als Eingabeende
		while (!(UCSRA & (1<<RXC))) {
			;    //wartet auf Eingabe
		}
		t=UDR;
		if (t==0x08) { // Backspace
			if (i<maxbits) { // i=11 ist groesster wert
				i++;
				uart_putc(t); // Echo der Eingabe
			}
		}
		else {
			if (i>0) { // Maximale Lnge n0ch nicht erreicht ?
				if (t=='1') {
					*zahl=*zahl|(1<<(i-1));
					uart_putc('1');
					i--;
				}
				if (t=='0') {
					*zahl=*zahl&~(1<<(i-1));
					uart_putc('0');
					i--;
				}
			}
		}
	}
}

void inp_time(volatile datestruct *time) // Zeiger auf zeit-sruct
{
	//Stunde
	uart_puttext(strcpy_P(string60, PSTR("Stunde:")));
	uart_puttext(strcpy_P(string60, PSTR("(")));
	strcpy(string15,"");
	itos(string15,time->hour,2);
	uart_puttext(string15); // string15 mit akt. Wert laden
	uart_puttext(strcpy_P(string60, PSTR("):")));
	inp_string(string15,3);
	time->hour=atoi(string15);
	uart_putcrlf();
	//Minute
	uart_puttext(strcpy_P(string60, PSTR("Minute:")));
	uart_puttext(strcpy_P(string60, PSTR("(")));
	strcpy(string15,"");
	itos(string15,time->min,2);
	uart_puttext(string15); // string15 mit akt. Wert laden
	uart_puttext(strcpy_P(string60, PSTR("):")));
	inp_string(string15,3);
	time->min=atoi(string15);
	uart_putcrlf();
	//Sekunde
	uart_puttext(strcpy_P(string60, PSTR("Sekunde:")));
	uart_puttext(strcpy_P(string60, PSTR("(")));
	strcpy(string15,"");
	itos(string15,time->sec,2);
	uart_puttext(string15); // string15 mit akt. Wert laden
	uart_puttext(strcpy_P(string60, PSTR("):")));
	inp_string(string15,3);
	time->sec=atoi(string15);
}

void out_time( datestruct time ) // gibt Zeit in string60 aus
{
	strcpy(string60,"");
	itos(string15,time.hour,2);
	strcat(string60,string15);
	strcat(string60, ":");
	itos(string15,time.min,2);
	strcat(string60,string15);
	strcat(string60, ":");
	itos(string15,time.sec,2);
	strcat(string60,string15);
}

void anzeigen( void ) // ruft konfig-menue auf
{
	char *wday_text[] = {"So","Mo","Di","Mi","Do","Fr","Sa"};
	uint8_t i;

	//Uptime
	uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR("Uptime:               ")));
	strcpy(string60,"");
	ltos(string15,uptime.day,5);
	strcat(string60,string15);
	strcat(string60, "d ");
	itos(string15,uptime.hour,2);
	strcat(string60,string15);
	strcat(string60, "h ");
	itos(string15,uptime.min,2);
	strcat(string60,string15);
	strcat(string60, "m ");
	itos(string15,uptime.sec,2);
	strcat(string60,string15);
	strcat(string60, "s ");
	uart_puttext(string60);
	//Datum
	uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR("Datum+Uhrzeit:        ")));
	strcpy(string60,"");
	strcat(string60,wday_text[zeit.wday]);
	strcat(string60," ");
	itos(string15,zeit.day,2);
	strcat(string60,string15);
	strcat(string60, ".");
	itos(string15,zeit.month,2);
	strcat(string60,string15);
	strcat(string60, ".");
	ltos(string15,zeit.year,4);
	strcat(string60,string15);
	uart_puttext(string60);
	//Uhrzeit
	// uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR(" ")));
	out_time(zeit);
	uart_puttext(string60);
	//Quarzkorektur
	uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR("Quarz-Korrektur:      ")));
	ltos(string60,quarz_adjustment,4);
	uart_puttext(string60);
	//Uhrzeit Mo-Fr
	uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR("Uhrzeit Mo-Fr hoch:   ")));
	out_time(hoch_mofr);
	uart_puttext(string60);
	//Uhrzeit
	//uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR(" runter:")));
	strcpy(string60,"");
	out_time(runter_mofr);
	uart_puttext(string60);
	//Uhrzeit
	uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR("Uhrzeit Sa+So hoch:   ")));
	out_time(hoch_saso);
	uart_puttext(string60);
	//Uhrzeit
	//uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR(" runter:")));
	out_time(runter_saso);
	uart_puttext(string60);
	//Zeit-enable
	uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR("Zeitsteuerung:        ")));
	itos(string60,zeitsteuer_enable,1);
	uart_puttext(string60);
	//zeitgesteuert-Tasten
	uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR("Zeitgest.-Tasten hoch:")));
	print_nibble(zeitgest_hoch_val,2);
	print_nibble(zeitgest_hoch_val,1);
	print_nibble(zeitgest_hoch_val,0);
	//uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR(" runter:")));
	print_nibble(zeitgest_runter_val,2);
	print_nibble(zeitgest_runter_val,1);
	print_nibble(zeitgest_runter_val,0);
	//Haltezeiten
	uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR("Haltezeiten:\n\r")));
	strcpy(string60,"");
	for (i=0;i<12;i++) {
		itos(string15,i,2);
		strcat(string60,string15);
		strcat(string60," ");
	}
	uart_puttext(string60);
	uart_putcrlf();
	strcpy(string60,"");
	for (i=0;i<12;i++) {
		itos(string15,time_tab_total[i],2);
		strcat(string60,string15);
		strcat(string60," ");
	}
	uart_puttext(string60);

	//Doppelklick
	uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR("Doppelklick:          ")));
	itos(string60,doppel_klick_val,3);
	uart_puttext(string60);
	//template_nr
	uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR("Doppelklick-Template: ")));
	itos(string60,template_nr,1);
	uart_puttext(string60);
	//Mapping
	uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR("Doppelklick-Mapping:\n\r")));
	for (i=0;i<12;i+=2) {
		uart_puttext(strcpy_P(string60, PSTR("Taste #")));
		itos(string60,i,2);
		uart_puttext(string60);
		uart_puttext(strcpy_P(string60, PSTR(" (hoch): ")));
		print_nibble(template_table[template_nr][i],2);
		print_nibble(template_table[template_nr][i],1);
		print_nibble(template_table[template_nr][i],0);
		uart_puttext(strcpy_P(string60, PSTR(" #")));
		itos(string60,i+1,2);
		uart_puttext(string60);
		uart_puttext(strcpy_P(string60, PSTR(" (runter): ")));
		print_nibble(template_table[template_nr][i+1],2);
		print_nibble(template_table[template_nr][i+1],1);
		print_nibble(template_table[template_nr][i+1],0);
		uart_putcrlf();
	}
	//auf Tastendruck warten
	uart_puttext(strcpy_P(string60, PSTR("\n\rTaste druecken")));
	while (!(UCSRA & (1<<RXC))) {
		;    //wartet auf Eingabe
	}
	i=UDR; // Dummy-Lesen
}


void fernsteuerung_rolladen( void ) // manuelle Fernbedieunung ber das Men
{
	char t;
	uint8_t exit=0;

	while (exit==0) {
		uart_putcrlf();
		uart_puttext(strcpy_P(string60, PSTR("Fernsteuerung:\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("A - 0 hoch\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("B - 0 runter\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("C - 1 hoch\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("D - 1 runter\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("E - 2 hoch\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("F - 2 runter\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("G - 3 hoch\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("H - 3 runter\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("I - 4 hoch\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("J - 4 runter\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("K - 5 hoch\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("L - 5 runter\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("M - Alle hoch\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("N - Alle runter\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("X - Menu Verlassen\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("Kommando eingeben:")));

		while (!(UCSRA & (1<<RXC))) {
			;
		}
		t=UDR;
		uart_putc(t);



// Matrix:
// 4095 = 111111111111 alle hoch
// 1365 = 010101010101 alle runter
// 
//
//
//

uint16_t map_rolladen_bits2( uint8_t rolladennummer, uint8_t richtung )
{
	uint16_t rolladen_bit_tab[2][6]={	{ 0x0003, 0x000c, 0x0030, 0x00c0, 0x0300, 0x0c00 },//hoch
										{ 0x0001, 0x0004, 0x0010, 0x0040, 0x0100, 0x0400 }	}; //runter

	return(rolladen_bit_tab[richtung][rolladennummer]);
}

	uint16_t rolladen_bits;


		switch (t) {
		case 'a'  :
			rolladen_bits=map_rolladen_bits2(0,0); //0=hoch
			switch_relais(rolladen_bits, output_ist);
			output_ist=rolladen_bits;
			exit=1;
			break;
		case 'b'  :
			rolladen_bits=map_rolladen_bits2(0,1); //1=runter
			switch_relais(rolladen_bits, output_ist);
			output_ist=rolladen_bits;
			exit=1;
			break;
		case 'c'  :
			rolladen_bits=map_rolladen_bits2(1,0); //0=hoch
			switch_relais(rolladen_bits, output_ist);
			output_ist=rolladen_bits;
			exit=1;
			break;
		case 'd'  :
			rolladen_bits=map_rolladen_bits2(1,1); //1=runter
			switch_relais(rolladen_bits, output_ist);
			output_ist=rolladen_bits;
			exit=1;
			break;
		case 'e'  :
			rolladen_bits=map_rolladen_bits2(2,0); //0=hoch
			switch_relais(rolladen_bits, output_ist);
			output_ist=rolladen_bits;
			exit=1;
			break;
		case 'f'  :
			rolladen_bits=map_rolladen_bits2(2,1); //1=runter
			switch_relais(rolladen_bits, output_ist);
			output_ist=rolladen_bits;
			exit=1;
			break;
		case 'g'  :
			rolladen_bits=map_rolladen_bits2(3,0); //0=hoch
			switch_relais(rolladen_bits, output_ist);
			output_ist=rolladen_bits;
			exit=1;
			break;
		case 'h'  :
			rolladen_bits=map_rolladen_bits2(3,1); //1=runter
			switch_relais(rolladen_bits, output_ist);
			output_ist=rolladen_bits;
			exit=1;
			break;
		case 'i'  :
			rolladen_bits=map_rolladen_bits2(4,0); //0=hoch
			switch_relais(rolladen_bits, output_ist);
			output_ist=rolladen_bits;
			exit=1;
			break;
		case 'j'  :
			rolladen_bits=map_rolladen_bits2(4,1); //1=runter
			switch_relais(rolladen_bits, output_ist);
			output_ist=rolladen_bits;
			exit=1;
			break;
		case 'k'  :
			rolladen_bits=map_rolladen_bits2(5,0); //0=hoch
			switch_relais(rolladen_bits, output_ist);
			output_ist=rolladen_bits;
			exit=1;
			break;
		case 'l'  :
			rolladen_bits=map_rolladen_bits2(5,1); //1=runter
			switch_relais(rolladen_bits, output_ist);
			output_ist=rolladen_bits;
			exit=1;
			break;
		case 'm'  :
			uart_putcrlf();
			switch_relais(4095, output_ist);
			output_ist=4095; //faehrt alle rolladen hoch
			exit=1;
			break;

		case 'n'  :
			uart_putcrlf();
			switch_relais(1365, output_ist);
			output_ist=1365; //faehrt alle rolladen runter
			exit=1;
			break;


		case 'x'  :
			uart_putcrlf();
			exit=1;
			break;
		}
		uart_putcrlf();
	}
}


void konfig_zeit( void ) // ruft zeit-menue auf
{
	char t;
	uint8_t exit=0;

	while (exit==0) {
		uart_putcrlf();
		uart_puttext(strcpy_P(string60, PSTR("Zeitinformationen etc. eingeben:\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("U - Uhrzeit\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("D - Datum\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("W - Schaltzeiten Mo.-Fr.\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("S - Schaltzeiten Sa.+So.\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("E - Zeitsteuerung ein/aus\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("M - Mapping fuer Zeitsteuerung\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("Q - Quarz-Korrektur\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("A - Anzeige der Konfiguration\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("X - Menu Verlassen\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("Kommando eingeben:")));

		while (!(UCSRA & (1<<RXC))) {
			;
		}
		t=UDR;
		uart_putc(t);

		switch (t) {
		case 'u'  :
			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Aktuelle Uhrzeit eingeben:\n\r")));
			inp_time(&zeit);
			break;

		case 'w'  :
			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Uhrzeit Rolladen oeffnen Mo.-Fr. eingeben:\n\r")));
			inp_time(&hoch_mofr);

			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Uhrzeit Rolladen schliessen Mo.-Fr. eingeben:\n\r")));
			inp_time(&runter_mofr);
			break;

		case 's'  :
			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Uhrzeit Rolladen oeffnen Sa.+So. eingeben:\n\r")));
			inp_time(&hoch_saso);

			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Uhrzeit Rolladen schliessen Sa.+So. eingeben:\n\r")));
			inp_time(&runter_saso);
			break;

		case 'd'  :
			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("aktuelles Datum eingeben:\n\r")));
			// Tag
			uart_puttext(strcpy_P(string60, PSTR("Tag:")));
			uart_puttext(strcpy_P(string60, PSTR("(")));
			strcpy(string15,"");
			itos(string15,zeit.day,2);
			uart_puttext(string15); // string15 mit akt. Wert laden
			uart_puttext(strcpy_P(string60, PSTR("):")));
			inp_string(string15,3);
			zeit.day=atoi(string15);
			//Monat
			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Monat:")));
			uart_puttext(strcpy_P(string60, PSTR("(")));
			strcpy(string15,"");
			itos(string15,zeit.month,2);
			uart_puttext(string15);// string15 mit akt. Wert laden
			uart_puttext(strcpy_P(string60, PSTR("):")));
			inp_string(string15,3);
			zeit.month=atoi(string15);
			//Jahr
			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Jahr:")));
			uart_puttext(strcpy_P(string60, PSTR("(")));
			strcpy(string15,"");
			ltos(string15,zeit.year,4);
			uart_puttext(string15);// string15 mit akt. Wert laden
			uart_puttext(strcpy_P(string60, PSTR("):")));
			inp_string(string15,5);
			zeit.year=atol(string15);
			//Wochentag
			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Wochentag (So=0, Mo=1, Di=2, Mi=3, Do=4, Fr=5, Sa=6):")));
			uart_puttext(strcpy_P(string60, PSTR("(")));
			strcpy(string15,"");
			itos(string15,zeit.wday,1);
			uart_puttext(string15);// string15 mit akt. Wert laden
			uart_puttext(strcpy_P(string60, PSTR("):")));
			inp_string(string15,2);
			zeit.wday=atoi(string15);
			break;

		case 'q'  :
			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Quarz-Korrektur:")));
			uart_puttext(strcpy_P(string60, PSTR("(")));
			strcpy(string15,"");
			ltos(string15,quarz_adjustment,4);
			uart_puttext(string15);// string15 mit akt. Wert laden
			uart_puttext(strcpy_P(string60, PSTR("):")));
			inp_string(string15,5);
			quarz_adjustment=atol(string15);
			break;


		case 'm'  :
			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Tastencode hoch:")));
			uart_puttext(strcpy_P(string60, PSTR("(")));
			print_nibble(zeitgest_hoch_val,2);
			print_nibble(zeitgest_hoch_val,1);
			print_nibble(zeitgest_hoch_val,0);
			uart_puttext(strcpy_P(string60, PSTR("):")));
			inp_bits(&zeitgest_hoch_val);

			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Tastencode runter:")));
			uart_puttext(strcpy_P(string60, PSTR("(")));
			print_nibble(zeitgest_runter_val,2);
			print_nibble(zeitgest_runter_val,1);
			print_nibble(zeitgest_runter_val,0);
			uart_puttext(strcpy_P(string60, PSTR("):")));
			inp_bits(&zeitgest_runter_val);
			break;

		case 'e'  :
			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Zeitsteuerung 0=disable, 1=enable:")));
			uart_puttext(strcpy_P(string60, PSTR("(")));
			strcpy(string15,"");
			itos(string15,zeitsteuer_enable,1);
			uart_puttext(string15);// string15 mit akt. Wert laden
			uart_puttext(strcpy_P(string60, PSTR("):")));
			inp_string(string15,2);
			zeitsteuer_enable=atoi(string15);
			break;

		case 'x'  :
			uart_putcrlf();
			exit=1;
			break;
		case 'a'  :
			anzeigen();
			break;
		}
		uart_putcrlf();
	}
}

uint16_t map_rolladen_bits( uint8_t rolladennummer, uint8_t richtung )
{
	uint16_t rolladen_bit_tab[2][6]={	{ 0x0003, 0x000c, 0x0030, 0x00c0, 0x0300, 0x0c00 },//hoch
										{ 0x0001, 0x0004, 0x0010, 0x0040, 0x0100, 0x0400 }	}; //runter

	return(rolladen_bit_tab[richtung][rolladennummer]);
}

void wizzard( void ) // Ermittelt Rolladen-laufzeiten runter und hoch fr vorgegebenen Rolladen
{
	uint8_t rolladen_nr,wizzard_sec_stop,wizzard_sec_start, wizzard_dt;
	uint16_t rolladen_bits;

	uart_putcrlf();

	//Rolladen auswhlen
	uart_puttext(strcpy_P(string60, PSTR("Rolladen auswaehlen (0-5):")));
	inp_string(string15,2);
	uart_putcrlf();

	//gewhlter Rolladen hoch fahren
	rolladen_nr=atoi(string15);
	rolladen_bits=map_rolladen_bits(rolladen_nr,0); //0=hoch
	switch_relais(rolladen_bits, output_ist);
	output_ist=rolladen_bits;

	//Auf Tastendruck warten
	uart_puttext(strcpy_P(string60, PSTR("Rolladentaste druecken wenn Rolladen hochgefahren ist\n\r")));
	real_keys_pressed=0; // abfrage initialisieren
	while (!real_keys_pressed) {
		;    //wartet auf Eingabe (Achtung: irgendeine Rolladentaste !!!!)
	}

	//Timer Starten+Rolladen nach unten fahren
	rolladen_bits=map_rolladen_bits(rolladen_nr,1); //1=runter
	wizzard_sec_start=zeit.sec;
	switch_relais(rolladen_bits, output_ist);
	output_ist=rolladen_bits;

	//auf tastendruck warten
	uart_puttext(strcpy_P(string60, PSTR("Rolladentaste druecken wenn Rolladen runtergefahren ist\n\r")));
	real_keys_pressed=0; // abfrage initialisieren
	while (!real_keys_pressed) {
		;    //wartet auf Eingabe (Achtung: irgendeine Rolladentaste !!!!)
	}

	//Timer+Rolladen stoppen
	switch_relais(0, output_ist);
	output_ist=0;
	wizzard_sec_stop=zeit.sec;

	//dt ermitteln und abspeichern
	if (wizzard_sec_stop<wizzard_sec_start) wizzard_sec_stop+=60;
	wizzard_dt=wizzard_sec_stop-wizzard_sec_start;
	time_tab_total[rolladen_nr*2]=wizzard_dt+1; //sicherheitshalber eine Sekunde zugeben !

	//auf tastendruck warten
	//real_keys_pressed=0; // abfrage initialisieren
	//while (!real_keys_pressed) { ;} //wartet auf Eingabe (Achtung: irgendeine Rolladentaste !!!!)

	//Timer Starten+Rolladen nach oben fahren
	rolladen_bits=map_rolladen_bits(rolladen_nr,0); //0=hoch
	wizzard_sec_start=zeit.sec;
	switch_relais(rolladen_bits, output_ist);
	output_ist=rolladen_bits;

	//auf tastendruck warten
	uart_puttext(strcpy_P(string60, PSTR("Rolladentaste druecken wenn Rolladen hochgefahren ist\n\r")));
	real_keys_pressed=0; // abfrage initialisieren
	while (!real_keys_pressed) {
		;    //wartet auf Eingabe (Achtung: irgendeine Rolladentaste !!!!)
	}
	//Timer stoppen
	switch_relais(0, output_ist);
	output_ist=0;
	wizzard_sec_stop=zeit.sec;

	//dt ermitteln und abspeich
	if (wizzard_sec_stop<wizzard_sec_start) wizzard_sec_stop+=60;
	wizzard_dt=wizzard_sec_stop-wizzard_sec_start;
	time_tab_total[rolladen_nr*2+1]=wizzard_dt+1;

	//Tastendruck verwerfen
	real_keys=4095;
	real_keys_old=real_keys;
}

void konfig_rolladen( void ) // ruft rolladen-menue auf
{
	char t;
	uint8_t i;
	uint8_t exit=0;

	while (exit==0) {
		t='#';
		uart_putcrlf();
		uart_puttext(strcpy_P(string60, PSTR("Rolladen-Parameter:\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("Z - Laufzeiten der Rolladen eingeben\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("W - Wizzard: Laufzeiten der Rolladen ermitteln\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("S - Doppelklick-Geschwindigkeit\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("T - Auswahl Doppelklick-Template\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("M - Editieren Doppelklick-Template #3\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("H - Alle Rolladen hoch\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("R - Alle Rolladen runter\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("A - Anzeige der Konfiguration\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("X - Menu verlassen\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("Kommando eingeben:")));

		while (!(UCSRA & (1<<RXC))) {
			;    // wartet auf Tastedruck
		}
		t=UDR;
		uart_putc(t);

		switch (t) {

		case 's'  :
			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Doppelklick-Geschwindigkeit(1-255):")));
			uart_puttext(strcpy_P(string60, PSTR("(")));
			strcpy(string15,"");
			itos(string15,doppel_klick_val,3);
			uart_puttext(string15);// string15 mit akt. Wert laden
			uart_puttext(strcpy_P(string60, PSTR("):")));
			inp_string(string15,4);
			doppel_klick_val=atoi(string15);
			break;

		case 'h'  :
			uart_putcrlf();
			switch_relais(4095, output_ist);
			output_ist=4095; //faehrt alle rolladen hoch
			uart_puttext(strcpy_P(string60, PSTR("Taste druecken")));
			while (!(UCSRA & (1<<RXC))) {
				;    //wartet auf Eingabe
			}
			i=UDR; // Dummy-Lesen
			break;

		case 'r'  :
			uart_putcrlf();
			switch_relais(1365, output_ist);
			output_ist=1365;// faert alle rolladen unter
			uart_puttext(strcpy_P(string60, PSTR("Taste druecken")));
			while (!(UCSRA & (1<<RXC))) {
				;    //wartet auf Eingabe
			}
			i=UDR; // Dummy-Lesen
			break;

		case 'z'  :
			uart_putcrlf();
			for (i=0;i<12;i++) {
				uart_puttext(strcpy_P(string60, PSTR("\n\rHaltezeit Taster ")));
				strcpy(string60,"");
				itos(string60,i,2);
				uart_puttext(string60);
				uart_puttext(strcpy_P(string60, PSTR(" eingeben (")));
				strcpy(string15,"");
				itos(string15,time_tab_total[i],2);
				uart_puttext(string15);// string15 mit akt. Wert laden
				uart_puttext(strcpy_P(string60, PSTR("):")));
				uart_putcrlf();
				inp_string(string15,3);
				time_tab_total[i]=atoi(string15);
			}
			break;

		case 'w'  :
			wizzard();
			break;

		case 't'  :
			uart_putcrlf();
			uart_puttext(strcpy_P(string60, PSTR("Template-Auswahl (0-3): ")));
			uart_puttext(strcpy_P(string60, PSTR("(")));
			strcpy(string15,"");
			itos(string15,template_nr,3);
			uart_puttext(string15);// string15 mit akt. Wert laden
			uart_puttext(strcpy_P(string60, PSTR("):")));
			inp_string(string15,2);
			template_nr=atoi(string15);
			break;

		case 'm'  :
			uart_putcrlf();
			for (i=0;i<12;i++) {
				uart_puttext(strcpy_P(string60, PSTR("Tastencode fuer Taster #")));
				strcpy(string60,"");
				itos(string60,i,2);
				uart_puttext(string60);
				uart_puttext(strcpy_P(string60, PSTR(" eingeben (")));
				print_nibble(template_table[3][i],2);
				print_nibble(template_table[3][i],1);
				print_nibble(template_table[3][i],0);
				uart_puttext(strcpy_P(string60, PSTR("):")));
				inp_bits(&template_table[3][i]);
				uart_putcrlf();
			}
			break;

		case 'x'  :
			exit=1;
			break;

		case 'a'  :
			anzeigen();
			break;
		}
		uart_putcrlf();
	}
}

void load_eeprom_konfig( void )
{
	quarz_adjustment=eeprom_read_word(&eeprom_quarz_adjustment);
	template_nr=eeprom_read_byte(&eeprom_template_nr);
	doppel_klick_val=eeprom_read_byte(&eeprom_doppel_klick_val);
	eeprom_read_block(time_tab_total,eeprom_time_tab_total, sizeof(eeprom_time_tab_total));

	//Zeitsteuerung
	eeprom_read_block((unsigned char*)&hoch_mofr,&eeprom_hoch_mofr, sizeof(eeprom_hoch_mofr));
	eeprom_read_block((unsigned char*)&runter_mofr,&eeprom_runter_mofr, sizeof(eeprom_runter_mofr));
	eeprom_read_block((unsigned char*)&hoch_saso,&eeprom_hoch_saso, sizeof(eeprom_hoch_saso));
	eeprom_read_block((unsigned char*)&runter_saso,&eeprom_runter_saso, sizeof(eeprom_runter_saso));
	zeitsteuer_enable=eeprom_read_byte(&eeprom_zeitsteuer_enable);
	zeitgest_hoch_val=eeprom_read_word(&eeprom_zeitgest_hoch_val);
	zeitgest_runter_val=eeprom_read_word(&eeprom_zeitgest_runter_val);

	// mapping-tab[3]
	eeprom_read_block(&template_table[3][0],eeprom_template_table_3, sizeof(eeprom_template_table_3));
}

void konfig( void ) // ruft konfig-menue auf
{
	char t;
	uint8_t exit=0;

	while (exit==0) {
		t='#';
		uart_putc(0x0c);
		uart_puttext(strcpy_P(string60, PSTR("Hauptmenue:\n\r")));
		uart_putcrlf();
		uart_puttext(strcpy_P(string60, PSTR("Z - Einstellung Zeit, Datum etc.\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("F - Fernsteuerung hoch/runter.\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("R - Einstellung der Rollaeden\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("A - Anzeigen der aktuellen Einstellungen\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("S - Konfiguration-Speichern\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("L - Konfiguration-Laden\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("X - Menu verlassen\n\r")));
		uart_puttext(strcpy_P(string60, PSTR("Kommando eingeben:")));

		while (!(UCSRA & (1<<RXC))) {
			;
		}
		t=UDR;
		uart_putc(t);

		switch (t) {
		case 'z'  :
			konfig_zeit();
			break;

		case 'r'  :
			konfig_rolladen();
			break;

		case 'f'  :
			fernsteuerung_rolladen();
			break;


		case 'a'  :
			anzeigen();
			break;

		case 's'  :	// Sichert ins EEPROM:
			eeprom_write_word(&eeprom_quarz_adjustment,quarz_adjustment);
			eeprom_write_byte(&eeprom_template_nr,template_nr);
			eeprom_write_byte(&eeprom_doppel_klick_val,doppel_klick_val);
			eeprom_write_block(time_tab_total,eeprom_time_tab_total, sizeof(eeprom_time_tab_total));

			//Zeitsteuerung
			eeprom_write_block((unsigned char*)&hoch_mofr,&eeprom_hoch_mofr, sizeof(eeprom_hoch_mofr));
			eeprom_write_block((unsigned char*)&runter_mofr,&eeprom_runter_mofr, sizeof(eeprom_runter_mofr));
			eeprom_write_block((unsigned char*)&hoch_saso,&eeprom_hoch_saso, sizeof(eeprom_hoch_saso));
			eeprom_write_block((unsigned char*)&runter_saso,&eeprom_runter_saso, sizeof(eeprom_runter_saso));
			eeprom_write_byte(&eeprom_zeitsteuer_enable,zeitsteuer_enable);
			eeprom_write_word(&eeprom_zeitgest_hoch_val,zeitgest_hoch_val);
			eeprom_write_word(&eeprom_zeitgest_runter_val,zeitgest_runter_val);

			// mapping-tab[3]
			eeprom_write_block(&template_table[3][0],eeprom_template_table_3, sizeof(eeprom_template_table_3));
			break;

		case 'l'  : // Ldt Konfig aus EEPROM
			load_eeprom_konfig();
			break;

		case 'x'  :
			exit=1;
			break;
		}
		uart_putcrlf();
	}
}

uint8_t compare(datestruct time1, datestruct time2) // vergleicht die aktuelle Zeit mit den Schaltzeiten
{
	uint8_t hcmp=0;
	uint8_t mcmp=0;
	uint8_t scmp=0;

	if (time1.hour == time2.hour) hcmp=1;
	if (time1.min == time2.min) mcmp=1;
	if (time1.sec == time2.sec) scmp=1;
	if ((hcmp==1)&&(mcmp==1)&&(scmp==1)) return(1); // 1 wenn Zeiten gleich sind
	return(0);
}

int main (void) {

	uint8_t old_sec;
	uint16_t v_keys, v_keys_old , v_keys_timer , zeitgesteuert;

	// Controller Konigurieren
	DDRA  = 0x00; // PA0-7 Eingnge
	DDRB  = 0xf0; // PB0-3 Eingnge
	DDRD  = 0xf0; // PD4-7 Ausgnge
	DDRC  = 0xff; // PC0-7 Ausgnge

	//Timer konfigurieren
	TCCR1A = 0x00;
	TCCR1B = 1<<WGM12 ^ 1<<CS10; // set CTC-Mode "4" + no prescaler
	OCR1A = T_VAL;
	TIMSK = 1<<OCIE1A;

	//UART konfigurieren
	UCSRB |= (1<<RXEN) | (1<<TXEN);    // UART TX+RX einschalten
	UCSRC |= (1<<URSEL)|(3<<UCSZ0);    // Asynchron 8N1
	UBRRH = BAUD >> 8;		            // Baud-Rate
	UBRRL = BAUD & 0xFF;

	//Versionsinfo
	uart_putcrlf();
	uart_putcrlf();
	uart_puttext(strcpy_P(string60, PSTR("Version vom: ")));
	uart_puttext(strcpy_P(string60, PSTR(__DATE__)));
	uart_puttext(strcpy_P(string60, PSTR(" - ")));
	uart_puttext(strcpy_P(string60, PSTR(__TIME__)));
	uart_putcrlf();

	//Variablen initialisieren
	output_ist=0;
	v_keys=4095;
	v_keys_old=4095;
	v_keys_timer=4095; //Initialisierung
	old_sec=zeit.sec;
	zeitgesteuert=0;

	//ldt gespeicherte Konfig-Werte
	//load_eeprom_konfig();

	//Interupts erlauben
	sei();

	//Hauptschleife
	while (1) {

		//CR-Taste startet Konfig-Mode
		if (UCSRA & (1<<RXC)) {
			if (UDR=='\r') {
				konfig();
			}
		}

		//Erzeugung 12-BIT-Zahl aus Eingangstasten wird in Interrupt-Routine erzeugt
		// real_keys enthlt Taste, Doppel enthlt die Doppelklicks

		v_keys=real_keys;
		if (doppel_klick!=0) {
			v_keys=map_keys(~doppel_klick); //doppel- berschreibt einzel-klick
			v_keys|=doppel_klick; // Blendet ursprnglich gedrckte Taste aus (sonst Stopp-Funktionfr diese Tase)
		}

		//zeitgesteuert
		if (zeitgesteuert!=0) {
			v_keys=zeitgesteuert;
			zeitgesteuert=0;
		}

		//Laufzeit pro Taster Setzen
		v_keys_timer=startstop_time(v_keys, v_keys_old, v_keys_timer);

		//Zeit fortschreiben
		if (zeit.sec!=old_sec) {
			v_keys_timer=continue_time(v_keys_timer);

			//Schaltzeiten fr zeitgesteuerten-Mode
			if (zeitsteuer_enable!=0) {
				if ((zeit.wday>0) && (zeit.wday<6)) { //Mo-Fr
					if (compare(zeit,hoch_mofr)==1) zeitgesteuert=zeitgest_hoch_val; // 101010101010 -> hier kann Rolladenmapping gemacht werden
					if (compare(zeit,runter_mofr)==1) zeitgesteuert=zeitgest_runter_val; // 010101010101 -> hier kann Rolladenmapping gemacht werden
				}
				else { // Sa+So
					if (compare(zeit,hoch_saso)==1) zeitgesteuert=zeitgest_hoch_val; // 101010101010 -> hier kann Rolladenmapping gemacht werden
					if (compare(zeit,runter_saso)==1) zeitgesteuert=zeitgest_runter_val; // 010101010101 -> hier kann Rolladenmapping gemacht werden
				}
			}

			old_sec=zeit.sec;
		}
		//Zuweisen der Ausgnge abh. von Tasten
		output_soll=keys2relais(v_keys_timer);

		//Relais-Schaltung ausfhren incl. Schonung der Relais
		switch_relais(output_soll, output_ist);

		//Ausgabe zur vollen Sekunde oder wenn Taste gedrckt wurde
		if ((zeit.sec != old_sec)||(v_keys!=v_keys_old)) {

			//gibt uhrzeit aus
			uart_puttext(strcpy_P(string60, PSTR(" Z:")));
			out_time(zeit);
			uart_puttext(string60);
			//gibt Datum aus
			uart_puttext(strcpy_P(string60, PSTR(" D:")));
			strcpy(string60,"");
			itos(string15,zeit.day,2);
			strcat(string60,string15);
			strcat(string60, ".");
			itos(string15,zeit.month,2);
			strcat(string60,string15);
			strcat(string60, ".");
			ltos(string15,zeit.year,4);
			strcat(string60,string15);
			uart_puttext(string60);
			//Ausgabe ders Virtualisierungs-Mappings
			uart_puttext(strcpy_P(string60, PSTR(" V:")));
			print_nibble(v_keys,2);
			print_nibble(v_keys,1);
			print_nibble(v_keys,0);
			uart_puttext(strcpy_P(string60, PSTR(" ->")));
			//Ausgabe der Timer
			uart_puttext(strcpy_P(string60, PSTR(" T:")));
			print_nibble(v_keys_timer,2);
			print_nibble(v_keys_timer,1);
			print_nibble(v_keys_timer,0);
			uart_puttext(strcpy_P(string60, PSTR(" ->")));
			//Ausgabe der Output-PINs
			uart_puttext(strcpy_P(string60, PSTR(" O:")));
			print_nibble(output_soll,2);
			print_nibble(output_soll,1);
			print_nibble(output_soll,0);
			// abschlussverarbeitung
			uart_putcrlf();
		}
		//soll=ist
		output_ist=output_soll;
		v_keys_old=v_keys;
	}
	return 0;
}

